(** @canonical Dune_re.Re *)
module Re = Dune_re__Re

(** @canonical Dune_re.Re_automata *)
module Re_automata = Dune_re__Re_automata

(** @canonical Dune_re.Re_cset *)
module Re_cset = Dune_re__Re_cset

(** @canonical Dune_re.Re_fmt *)
module Re_fmt = Dune_re__Re_fmt
